#include "ScoutManager.h"
#include "ProductionManager.h"
#include "AttackManager.h"

using namespace UAlbertaBot;

ScoutManager::ScoutManager()
	: _workerScout(nullptr)
	, _overlordScout(nullptr)
	, _scoutStatus("None")
	, _gasStealStatus("None")
	, _scoutLocationOnly(false)
	, _scoutUnderAttack(false)
	, _tryGasSteal(false)
	, _didGasSteal(false)
	, _gasStealFinished(false)
	, _currentRegionVertexIndex(-1)
	, _previousScoutHP(0)
{
}

ScoutManager & ScoutManager::Instance()
{
	static ScoutManager instance;
	return instance;
}

void ScoutManager::update()
{
	// If we're not scouting now, minimum effort.
	if (!_workerScout && !_overlordScout)
	{
		return;
	}

	if (_workerScout)
	{
		// If the worker scout is gone, admit it.
		if (!_workerScout->exists() || _workerScout->getHitPoints() <= 0 ||   // it died
			_workerScout->getPlayer() != BWAPI::Broodwar->self())             // it got mind controlled!	
		{
			_workerScout = nullptr;
			
		}
	}

	if (_overlordScout)
	{
		// If the overlord scout is gone, admit it.
		if 	(!_overlordScout->exists() || _overlordScout->getHitPoints() <= 0 ||  // it died
			_overlordScout->getPlayer() != BWAPI::Broodwar->self())                //  it got control

		{
			_overlordScout = nullptr;
			
		}
	}

	


	// If we only want to locate the enemy base and we have, release the scout worker.
	if (_scoutLocationOnly && InformationManager::Instance().getEnemyMainBaseLocation() && _workerScout)
	{
		releaseWorkerScout();
		return;
	}

	// calculate enemy region vertices if we haven't yet
	if (_enemyRegionVertices.empty())
	{
		calculateEnemyRegionVertices();
	}

	moveWorkerScout();
	moveOverlordScout();
}

void ScoutManager::setWorkerScout(BWAPI::Unit unit)
{
	// if we have a previous worker scout, release it back to the worker manager
	if (_workerScout)
	{
		WorkerManager::Instance().finishedWithWorker(_workerScout);
	}

	_workerScout = unit;
	WorkerManager::Instance().setScoutWorker(_workerScout);
}

void ScoutManager::setOverlordScout(BWAPI::Unit unit)
{
	if (_overlordScout)
		return;
	_overlordScout = unit;
}

// Send the scout home.
void ScoutManager::releaseWorkerScout()
{
	if (_workerScout)
	{
		WorkerManager::Instance().finishedWithWorker(_workerScout);
		_workerScout = nullptr;
	}
}

void ScoutManager::setGasSteal()
{
	_tryGasSteal = true;
}

void ScoutManager::setScoutLocationOnly()
{
	_scoutLocationOnly = true;
}

void ScoutManager::drawScoutInformation(int x, int y)
{

	BWAPI::Broodwar->drawTextScreen(x, y, "ScoutInfo: %s", _scoutStatus.c_str());
	BWAPI::Broodwar->drawTextScreen(x, y + 10, "GasSteal: %s", _gasStealStatus.c_str());
	for (size_t i(0); i < _enemyRegionVertices.size(); ++i)
	{
		BWAPI::Broodwar->drawCircleMap(_enemyRegionVertices[i], 4, BWAPI::Colors::Green, false);
		BWAPI::Broodwar->drawTextMap(_enemyRegionVertices[i], "%d", i);
	}
}


void ScoutManager::moveOverlordScout()
{
	if (!_overlordScout)
		return;

	int scoutHP = _overlordScout->getHitPoints() + _overlordScout->getShields();

	// get the enemy base location, if we have one
	// Note: In case of an enemy proxy or weird map, this might be our own base. Roll with it.
	BWTA::BaseLocation *enemyBase = InformationManager::Instance().getEnemyMainBaseLocation();

	// for each start location in the level
	if (!enemyBase)
	{
		_scoutStatus = "Enemy base unknown, exploring";

		BWTA::BaseLocation *destLocation = nullptr;
		static BWTA::BaseLocation *lastDestLocation = nullptr;
		double minDistance = -1;
		double distance;
		//BWAPI::Position overlord_position = AttackManager::Instance().getOverlordSurveyPosition();
		for (BWTA::BaseLocation * startLocation : BWTA::getStartLocations())
		{
			// if we haven't explored it yet
			if (!BWAPI::Broodwar->isExplored(startLocation->getTilePosition()))
			{
				distance = InformationManager::Instance().getMyMainBaseLocation()->getAirDistance(startLocation);
				if (minDistance == -1 || distance < minDistance)
				{
					minDistance = distance;
					destLocation = startLocation;
				}

			}
		}
		if (destLocation)
		{

			if (lastDestLocation && destLocation != lastDestLocation)
			{
				double minDistance = -1;
				double distance;
				//BWAPI::Position overlord_position = AttackManager::Instance().getOverlordSurveyPosition();
				for (BWTA::BaseLocation * startLocation : BWTA::getStartLocations())
				{
					// if we haven't explored it yet
					if (!BWAPI::Broodwar->isExplored(startLocation->getTilePosition()))
					{
						distance = lastDestLocation->getAirDistance(startLocation);
						if (minDistance == -1 || distance < minDistance)
						{
							minDistance = distance;
							destLocation = startLocation;
						}

					}
				}
			}
			// assign a unit to go scout it, for not in a same place like overlord.
			Micro::SmartMove(_overlordScout, BWAPI::Position(destLocation->getTilePosition()));
			lastDestLocation = destLocation;
			return;
		}
	}
	else
	{
		if (BWAPI::Broodwar->enemy()->getRace() != BWAPI::Races::Terran)
			// to stop in the enemy space.
			Micro::SmartMove(_overlordScout, BWAPI::Position(enemyBase->getTilePosition()));
		else
			// back move in startlocation
			Micro::SmartMove(_overlordScout, BWAPI::Position(BWAPI::Broodwar->self()->getStartLocation()));
	}


	_previousScoutHP = scoutHP;
}

void ScoutManager::moveWorkerScout()
{
	if (!_workerScout)
		return;

	int scoutHP = _workerScout->getHitPoints() + _workerScout->getShields();

	gasSteal();

	// get the enemy base location, if we have one
	// Note: In case of an enemy proxy or weird map, this might be our own base. Roll with it.
	BWTA::BaseLocation *enemyBase = InformationManager::Instance().getEnemyMainBaseLocation();

	int scoutDistanceThreshold = 30;

	if (_workerScout->isCarryingGas())
	{
		BWAPI::Broodwar->drawCircleMap(_workerScout->getPosition(), 10, BWAPI::Colors::Purple, true);
	}

	// if we initiated a gas steal and the worker isn't idle, 
	bool finishedConstructingGasSteal = _workerScout->isIdle() || _workerScout->isCarryingGas();
	if (!_gasStealFinished && _didGasSteal && !finishedConstructingGasSteal)
	{
		return;
	}
	// check to see if the gas steal is completed
	else if (_didGasSteal && finishedConstructingGasSteal)
	{
		_gasStealFinished = true;
	}

	// if we know where the enemy region is and where our scout is
	if (_workerScout && enemyBase)
	{
		int scoutDistanceToEnemy = MapTools::Instance().getGroundDistance(_workerScout->getPosition(), enemyBase->getPosition());
		bool scoutInRangeOfenemy = scoutDistanceToEnemy <= scoutDistanceThreshold;

		// we only care if the scout is under attack within the enemy region
		// this ignores if their scout worker attacks it on the way to their base
		if (scoutHP < _previousScoutHP)
		{
			_scoutUnderAttack = true;
		}

		if (!_workerScout->isUnderAttack() && !enemyWorkerInRadius())
		{
			_scoutUnderAttack = false;
		}

		// if the scout is in the enemy region
		if (scoutInRangeOfenemy)
		{
			// if the worker scout is not under attack
			if (!_scoutUnderAttack)
			{
				// get the closest enemy worker
				BWAPI::Unit closestWorker = enemyWorkerToHarass();

				// if there is a worker nearby, harass it
				if ( (!_tryGasSteal || _gasStealFinished) && closestWorker && (_workerScout->getDistance(closestWorker) < 800))
				{
					_scoutStatus = "Harass enemy worker";
					_currentRegionVertexIndex = -1;
					Micro::SmartAttackUnit(_workerScout, closestWorker);
				}
				// otherwise keep moving to the enemy region
				else
				{
					_scoutStatus = "Following perimeter";
					followPerimeter();
				}

			}
			// if the worker scout is under attack
			else
			{
				_scoutStatus = "Under attack inside, fleeing";
				followPerimeter();
			}
		}
		// if the scout is not in the enemy region
		else if (_scoutUnderAttack)
		{
			_scoutStatus = "Under attack inside, fleeing";

			followPerimeter();
		}
		else
		{
			_scoutStatus = "Enemy region known, going there";

			// move to the enemy region
			followPerimeter();
		}

	}

	// for each start location in the level
	if (!enemyBase)
	{
		_scoutStatus = "Enemy base unknown, exploring";

		BWTA::BaseLocation *destLocation = nullptr;
		static BWTA::BaseLocation *lastDestLocation = nullptr;
		double minDistance = -1;
		double distance;
		//BWAPI::Position overlord_position = AttackManager::Instance().getOverlordSurveyPosition();
		for (BWTA::BaseLocation * startLocation : BWTA::getStartLocations())
		{
			// if we haven't explored it yet
			if (!BWAPI::Broodwar->isExplored(startLocation->getTilePosition()) )
			{
				distance = InformationManager::Instance().getMyMainBaseLocation()->getAirDistance(startLocation);
				if (minDistance == -1 || distance < minDistance)
				{
					minDistance = distance;
					destLocation = startLocation;
				}
					
			}
		}
		if (destLocation)
		{

			if (lastDestLocation && destLocation != lastDestLocation)
			{
				double minDistance = -1;
				double distance;
				//BWAPI::Position overlord_position = AttackManager::Instance().getOverlordSurveyPosition();
				for (BWTA::BaseLocation * startLocation : BWTA::getStartLocations())
				{
					// if we haven't explored it yet
					if (!BWAPI::Broodwar->isExplored(startLocation->getTilePosition()))
					{
						distance = lastDestLocation->getAirDistance(startLocation);
						if (minDistance == -1 || distance < minDistance)
						{
							minDistance = distance;
							destLocation = startLocation;
						}

					}
				}
			}
			// assign a unit to go scout it, for not in a same place like overlord.
			Micro::SmartMove(_workerScout, BWAPI::Position(destLocation->getTilePosition()));
			lastDestLocation = destLocation;
			return;
		}

	}
	

	_previousScoutHP = scoutHP;
}

void ScoutManager::followPerimeter()
{
	BWAPI::Position fleeTo = getFleePosition();

	Micro::SmartMove(_workerScout, fleeTo);
}

void ScoutManager::gasSteal()
{
	if (!_tryGasSteal)
	{
		_gasStealStatus = "Not using gas steal";
		return;
	}

	if (_didGasSteal)
	{
		return;
	}

	if (!_workerScout)
	{
		_gasStealStatus = "No worker scout";
		return;
	}

	BWTA::BaseLocation  *enemyBase = InformationManager::Instance().getEnemyMainBaseLocation();
	if (!enemyBase)
	{
		_gasStealStatus = "No enemy base location found";
		return;
	}

	BWAPI::Unit enemyGeyser = getEnemyGeyser();
	if (!enemyGeyser)
	{
		_gasStealStatus = "Not exactly 1 enemy geyser";
		return;
	}

	if (!_didGasSteal)
	{
		ProductionManager::Instance().queueGasSteal();
		_didGasSteal = true;
		Micro::SmartMove(_workerScout, enemyGeyser->getPosition());
		_gasStealStatus = "Stealing gas";
	}
}

// Choose an enemy worker to harass, or none.
BWAPI::Unit ScoutManager::enemyWorkerToHarass()
{
	BWAPI::Unit enemyWorker = nullptr;
	double maxDist = 0;

	BWAPI::Unit geyser = getEnemyGeyser();

	for (auto & unit : BWAPI::Broodwar->enemy()->getUnits())
	{
		if (unit->getType().isWorker() && unit->isConstructing())
		{
			return unit;
		}
	}

	// Only harass terran SCVs that are building, no other workers.
	// Conclusion: Doesn't help. Scout dies if counterattacked at all.
	// return nullptr;

	// for each enemy worker
	for (auto & unit : BWAPI::Broodwar->enemy()->getUnits())
	{
		if (unit->getType().isWorker())
		{
			double dist = unit->getDistance(geyser);

			if (dist < 800 && dist > maxDist)
			{
				maxDist = dist;
				enemyWorker = unit;
			}
		}
	}

	return enemyWorker;
}

// If there is exactly 1 geyser in the enemy base, return it.
// If there's 0 we can't steal it, and if >1 then it's no use to steal it.
BWAPI::Unit ScoutManager::getEnemyGeyser()
{
	BWTA::BaseLocation *enemyBase = InformationManager::Instance().getEnemyMainBaseLocation();
	// 10 * 32 is evaluated value.
	BWAPI::Unitset geysers = BWAPI::Broodwar->getUnitsInRadius(enemyBase->getPosition(), 10 * 32, BWAPI::Filter::IsRefinery);
	if (geysers.size() == 1)
	{
		return *(geysers.begin());
	}

	return nullptr;
}

bool ScoutManager::enemyWorkerInRadius()
{
	for (auto & unit : BWAPI::Broodwar->enemy()->getUnits())
	{
		if (unit->getType().isWorker() && (unit->getDistance(_workerScout) < 300))
		{
			return true;
		}
	}

	return false;
}

// Unused.
bool ScoutManager::immediateThreat()
{
	BWAPI::Unitset enemyAttackingWorkers;
	for (auto & unit : BWAPI::Broodwar->enemy()->getUnits())
	{
		if (unit->getType().isWorker() && unit->isAttacking())
		{
			enemyAttackingWorkers.insert(unit);
		}
	}

	if (_workerScout->isUnderAttack())
	{
		return true;
	}

	for (auto & unit : BWAPI::Broodwar->enemy()->getUnits())
	{
		double dist = unit->getDistance(_workerScout);
		double range = unit->getType().groundWeapon().maxRange();

		if (unit->getType().canAttack() && !unit->getType().isWorker() && (dist <= range + 32))
		{
			return true;
		}
	}

	return false;
}

int ScoutManager::getClosestVertexIndex(BWAPI::Unit unit)
{
	int closestIndex = -1;
	double closestDistance = 10000000;

	for (size_t i(0); i < _enemyRegionVertices.size(); ++i)
	{
		double dist = unit->getDistance(_enemyRegionVertices[i]);
		if (dist < closestDistance)
		{
			closestDistance = dist;
			closestIndex = i;
		}
	}

	return closestIndex;
}

BWAPI::Position ScoutManager::getFleePosition()
{
	UAB_ASSERT_WARNING(!_enemyRegionVertices.empty(), "should have enemy region vertices");

	// if this is the first flee, we will not have a previous perimeter index
	if (_currentRegionVertexIndex == -1)
	{
		// so return the closest position in the polygon
		int closestPolygonIndex = getClosestVertexIndex(_workerScout);

		UAB_ASSERT_WARNING(closestPolygonIndex != -1, "Couldn't find a closest vertex");

		if (closestPolygonIndex == -1)
		{
			return BWAPI::Position(BWAPI::Broodwar->self()->getStartLocation());
		}
		else
		{
			// set the current index so we know how to iterate if we are still fleeing later
			_currentRegionVertexIndex = closestPolygonIndex;
			return _enemyRegionVertices[closestPolygonIndex];
		}
	}
	// if we are still fleeing from the previous frame, get the next location if we are close enough
	else
	{
		double distanceFromCurrentVertex = _enemyRegionVertices[_currentRegionVertexIndex].getDistance(_workerScout->getPosition());

		// keep going to the next vertex in the perimeter until we get to one we're far enough from to issue another move command
		while (distanceFromCurrentVertex < 128)
		{
			_currentRegionVertexIndex = (_currentRegionVertexIndex + 1) % _enemyRegionVertices.size();

			distanceFromCurrentVertex = _enemyRegionVertices[_currentRegionVertexIndex].getDistance(_workerScout->getPosition());
		}

		return _enemyRegionVertices[_currentRegionVertexIndex];
	}

}

void ScoutManager::calculateEnemyRegionVertices()
{
	BWTA::BaseLocation *enemyBase = InformationManager::Instance().getEnemyMainBaseLocation();


	if (!enemyBase)
	{
		return;
	}

	BWTA::Region * enemyRegion = BWTA::getRegion(enemyBase->getPosition());

	if (!enemyRegion)
	{
		return;
	}

	const BWAPI::Position basePosition = BWAPI::Position(BWAPI::Broodwar->self()->getStartLocation());
	const std::vector<BWAPI::TilePosition> & closestTobase = MapTools::Instance().getClosestTilesTo(basePosition);

	std::set<BWAPI::Position> unsortedVertices;

	// check each tile position
	for (size_t i(0); i < closestTobase.size(); ++i)
	{
		const BWAPI::TilePosition & tp = closestTobase[i];

		if (BWTA::getRegion(tp) != enemyRegion)
		{
			continue;
		}

		// a tile is 'surrounded' if
		// 1) in all 4 directions there's a tile position in the current region
		// 2) in all 4 directions there's a buildable tile
		bool surrounded = true;
		if (BWTA::getRegion(BWAPI::TilePosition(tp.x + 1, tp.y)) != enemyRegion || !BWAPI::Broodwar->isBuildable(BWAPI::TilePosition(tp.x + 1, tp.y))
			|| BWTA::getRegion(BWAPI::TilePosition(tp.x, tp.y + 1)) != enemyRegion || !BWAPI::Broodwar->isBuildable(BWAPI::TilePosition(tp.x, tp.y + 1))
			|| BWTA::getRegion(BWAPI::TilePosition(tp.x - 1, tp.y)) != enemyRegion || !BWAPI::Broodwar->isBuildable(BWAPI::TilePosition(tp.x - 1, tp.y))
			|| BWTA::getRegion(BWAPI::TilePosition(tp.x, tp.y - 1)) != enemyRegion || !BWAPI::Broodwar->isBuildable(BWAPI::TilePosition(tp.x, tp.y - 1)))
		{
			surrounded = false;
		}

		// push the tiles that aren't surrounded
		if (!surrounded && BWAPI::Broodwar->isBuildable(tp))
		{
			unsortedVertices.insert(BWAPI::Position(tp) + BWAPI::Position(16, 16));
		}
	}

	std::vector<BWAPI::Position> sortedVertices;
	BWAPI::Position current = *unsortedVertices.begin();

	_enemyRegionVertices.push_back(current);
	unsortedVertices.erase(current);

	// while we still have unsorted vertices left, find the closest one remaining to current
	while (!unsortedVertices.empty())
	{
		double bestDist = 1000000;
		BWAPI::Position bestPos;

		for (const BWAPI::Position & pos : unsortedVertices)
		{
			double dist = pos.getDistance(current);

			if (dist < bestDist)
			{
				bestDist = dist;
				bestPos = pos;
			}
		}

		current = bestPos;
		sortedVertices.push_back(bestPos);
		unsortedVertices.erase(bestPos);
	}

	// let's close loops on a threshold, eliminating death grooves
	int distanceThreshold = 100;

	while (true)
	{
		// find the largest index difference whose distance is less than the threshold
		int maxFarthest = 0;
		int maxFarthestStart = 0;
		int maxFarthestEnd = 0;

		// for each starting vertex
		for (int i(0); i < (int)sortedVertices.size(); ++i)
		{
			int farthest = 0;
			int farthestIndex = 0;

			// only test half way around because we'll find the other one on the way back
			for (size_t j(1); j < sortedVertices.size() / 2; ++j)
			{
				int jindex = (i + j) % sortedVertices.size();

				if (sortedVertices[i].getDistance(sortedVertices[jindex]) < distanceThreshold)
				{
					farthest = j;
					farthestIndex = jindex;
				}
			}

			if (farthest > maxFarthest)
			{
				maxFarthest = farthest;
				maxFarthestStart = i;
				maxFarthestEnd = farthestIndex;
			}
		}

		// stop when we have no long chains within the threshold
		if (maxFarthest < 4)
		{
			break;
		}

		double dist = sortedVertices[maxFarthestStart].getDistance(sortedVertices[maxFarthestEnd]);

		std::vector<BWAPI::Position> temp;

		for (size_t s(maxFarthestEnd); s != maxFarthestStart; s = (s + 1) % sortedVertices.size())
		{
			temp.push_back(sortedVertices[s]);
		}

		sortedVertices = temp;
	}

	_enemyRegionVertices = sortedVertices;
}